package de.lluni.javann.layers;

import de.lluni.javann.Initializer;
import de.lluni.javann.util.Utilities;
import org.ejml.simple.SimpleMatrix;

import java.util.Random;

/**
 * Fully connected layer with n Neurons
 */
public class FCLayer extends Layer {
    private final Initializer weightInit;
    private final Initializer biasInit;
    private SimpleMatrix weights;
    private SimpleMatrix biases;
    private int numNeurons;
    private boolean isInitialized;

    /**
     * Creates a fully connected layer with numNeurons neurons
     * @param numNeurons amount of neurons in this layer
     */
    public FCLayer(int numNeurons, Initializer weightInit, Initializer biasInit) {
        this.numNeurons = numNeurons;
        this.weightInit = weightInit;
        this.biasInit = biasInit;
        isInitialized = false;
    }

    private void initialize(int inputSize) {
        switch (weightInit) {
            case ZEROS -> this.weights = new SimpleMatrix(inputSize, numNeurons);
            case ONES -> this.weights = Utilities.ones(inputSize, numNeurons);
            case GAUSSIAN -> this.weights = Utilities.gaussianMatrix(inputSize, numNeurons, 0, 1, 0.1d);
            case RANDOM -> this.weights = Utilities.randomMatrix(inputSize, numNeurons);
        }
        switch (biasInit) {
            case ZEROS -> this.biases = new SimpleMatrix(1, numNeurons);
            case ONES -> this.biases = Utilities.ones(1, numNeurons);
            case GAUSSIAN -> this.biases = Utilities.gaussianMatrix(1, numNeurons, 0, 1, 0.1d);
            case RANDOM -> this.biases = Utilities.randomMatrix(1, numNeurons);
        }
        this.isInitialized = true;
    }

    /**
     * Updates input size when previous layer has newly added neurons.
     * @param n amount of new neurons in previous layer
     */
    public void updateInputSize(int n) {
        if (isInitialized) {
            Random random = new Random();

            // add new weights
            SimpleMatrix newWeights = new SimpleMatrix(this.weights.numRows() + n, this.weights.numCols());
            for (int i = 0; i < this.weights.numRows(); i++) {
                for (int j = 0; j < this.weights.numCols(); j++) {
                    newWeights.set(i, j, this.weights.get(i, j));
                }
            }
            for (int i = 0; i < newWeights.getNumElements(); i++) {
                if (newWeights.get(i) == 0) {
                    newWeights.set(i, 0.1d * random.nextGaussian(0, 1));
                }
            }
            this.weights = newWeights;
        }
    }

    /**
     * Adds new neurons at the end of the layer
     * @param n amount how many new neurons should be added
     */
    public void addNeuron(int n) {
        Random random = new Random();

        // update neuron count
        this.numNeurons += n;

        if (isInitialized) {
            // add new weights
            SimpleMatrix newWeights = new SimpleMatrix(this.weights.numRows(), this.weights.numCols() + n);
            for (int i = 0; i < this.weights.numRows(); i++) {
                for (int j = 0; j < this.weights.numCols(); j++) {
                    newWeights.set(i, j, this.weights.get(i, j));
                }
            }
            for (int i = 0; i < newWeights.getNumElements(); i++) {
                if (newWeights.get(i) == 0) {
                    newWeights.set(i, random.nextDouble(-1, 1));
                }
            }
            this.weights = newWeights;

            // add new biases
            SimpleMatrix newBiases = new SimpleMatrix(1, this.biases.numCols() + n);
            double[] newBiasValues = random.doubles(n, -1, 1).toArray();
            System.arraycopy(this.biases.getDDRM().data, 0, newBiases.getDDRM().data, 0, this.biases.numCols());
            System.arraycopy(newBiasValues, 0, newBiases.getDDRM().data, this.biases.numCols(), n);
            this.biases = newBiases;
        }
    }

    @Override
    public SimpleMatrix forwardPropagation(SimpleMatrix inputs) {
        if (!isInitialized) {
            initialize(inputs.numCols());
        }

        this.input = inputs;
        this.output = this.input.mult(this.weights).plus(this.biases);
        return this.output;
    }

    @Override
    public SimpleMatrix backwardPropagation(SimpleMatrix outputError, double learningRate) {
        SimpleMatrix inputError = outputError.mult(this.weights.transpose());
        SimpleMatrix weightsError = this.input.transpose().mult(outputError);

        this.weights = this.weights.plus(learningRate, weightsError);
        this.biases = this.biases.plus(learningRate, outputError);
        return inputError;
    }
}
